<'
# Copyright (c) 2011-2015,2017 LAAS/CNRS
# All rights reserved.
#
# Redistribution and use  in source  and binary  forms,  with or without
# modification, are permitted provided that the following conditions are
# met:
#
#   1. Redistributions of  source  code must retain the  above copyright
#      notice and this list of conditions.
#   2. Redistributions in binary form must reproduce the above copyright
#      notice and  this list of  conditions in the  documentation and/or
#      other materials provided with the distribution.
#
#                                      Anthony Mallet on Mon Mar  5 2012
#

if {[llength $argv] != 2} { error "expected arguments: component task" }
lassign $argv component task

# compute handy shortcuts
set comp [$component name]
set tname [$task name]

lang c
'>
#include "autoconf/acheader.h"

#include <sys/time.h>

#include <stdio.h>

#include "commonStructLib.h"
#include "h2evnLib.h"
<'if {![catch {$task period}]} {'>
#include "h2timerLib.h"
<'}'>
#include "taskLib.h"

#include "<"$comp">_control_task.h"
#include "<"$comp">_spinlock.h"


/* --- local data ---------------------------------------------------------- */

/* maximum <"$tname"> mbox size */
static __inline__ size_t
<"$comp">_<"$tname">_mbox_size(void)
{
<'  set n 0 '>
<'  foreach c [$task codels] {'>
<'    foreach p [$c parameters remote] {'>
<'      if {[incr n] == 1} {'>
  size_t s, m = 0;
<'      }'>
  s = genom_<"$comp">_<"[[$p remote] name]">_maxreply_size();
  if (m < s) m = s;
<'    }'>
<'  }'>
<'  foreach s [$task services] {'>
<'    foreach c [$s codels] {'>
<'      foreach p [$c parameters remote] {'>
<'      if {[incr n] == 1} {'>
  size_t s, m = 0;
<'      }'>
  s = genom_<"$comp">_<"[[$p remote] name]">_maxreply_size();
  if (m < s) m = s;
<'      }'>
<'    }'>
<'  }'>
<'  if {$n} {'>
  return (MBOX_BUF_SIZE(m)+MBOX_BUF_SIZE(sizeof(int)))*CLIENT_NMAX_RQST_ID;
<'  } else {'>
  return 0;
<'  }'>
}


/* <"[--- genom_${comp}_${tname}_exec_task ------------------------------]"> */

void *
genom_<"$comp">_<"$tname">_exec_task(void *data)
{
  enum {
    GENOM_SCHED_DONE = 1,
    GENOM_SCHED_ONE = 2,
    GENOM_SCHED_CONT = 4,
    GENOM_SCHED_WAKE = 8
  };
  struct genom_component_data *self = data;
  genom_state_rusage *usage = &self->tasks.<"[$task name]">.rusage;
  struct genom_activity *a;
  struct timeval tvs, tve;
<'if {![catch {$task period}]} {'>
  H2TIMER_ID timer;
<'}'>
  enum genom_activity_status e;
  int sched;
  size_t id;

<'if {![catch {$task period} period]} {'>
<'
  set ticks [expr {[$period value]/[[$component clock-rate] value]}]
  if {abs($ticks - round($ticks)) > 1e-6} {
    error [concat \
               "task $tname period ([$period value]) not a multiple of" \
               "clock-rate ([[$component clock-rate] value])"]
  }
  set ticks [expr round($ticks)]

  if {[catch {[$task delay] value} delay]} { set delay 0 }
  set delay [expr {$delay/[[$component clock-rate] value]}]
  if {abs($delay - round($delay)) > 1e-6} {
    error [concat \
               "task $tname delay ($delay) not a multiple of" \
               "clock-rate ([[$component clock-rate] value])"]
  }
  set delay [expr round($delay)]
'>
  /* create an h2 timer */
  timer = h2timerAlloc();
  if (timer == NULL) {
    genom_log_warn(1, "cannot create timer for %s exec task", "<"$tname">");
    goto signal_parent;
  }

  if (h2timerStart(timer, <"$ticks">, <"$delay">) != OK) {
    genom_log_warn(1, "cannot start timer for %s exec task", "<"$tname">");
    goto signal_parent;
  }
<'}'>

  /* initialize csLib for remotes */
  if (<"$comp">_<"$tname">_mbox_size()) {
    char mbox[64];
    snprintf(mbox, sizeof(mbox), "%s/%s", genom_instance, "<"$tname">");

    if (csMboxInit(mbox, 0, <"$comp">_<"$tname">_mbox_size()) == ERROR) {
      genom_log_warn(1, "cannot create %s task mailbox %s",
                     "<"$tname">", mbox);
      goto signal_parent;
    }
  }

  /* run */
  genom_log_info("spawned task <"[$task name]">");
  sched = GENOM_SCHED_CONT;
  gettimeofday(&tvs, NULL);
  tve = tvs;

  do {
    /* wait for next event */
<'if {[catch {$task period}]} {'>
    pthread_mutex_lock(&self->tasks.<"[$task name]">.lock);
    if (!(sched & GENOM_SCHED_CONT)) {
      struct timespec tout = { .tv_sec = tvs.tv_sec + 1,
                               .tv_nsec = tvs.tv_usec * 1000 };
      while (!self->tasks.<"[$task name]">.runnable)
        if (pthread_cond_timedwait(&self->tasks.<"[$task name]">.sync,
                                   &self->tasks.<"[$task name]">.lock,
                                   &tout))
          break;
    }
    genom_log_debug("<"[$task name]"> tick");
    gettimeofday(&tvs, NULL);
<'} else {'>
    if (!(sched & GENOM_SCHED_CONT)) {
      if (h2timerPause(timer) != OK) abort();
      gettimeofday(&tvs, NULL);
      genom_log_debug("<"[$task name]"> tick");
    }
    pthread_mutex_lock(&self->tasks.<"$tname">.lock);
    /* paused activities are only awaken on a period tick */
    self->tasks.<"[$task name]">.runnable = !(sched & GENOM_SCHED_CONT);
<'}'>
    sched = 0;

    if (self->tasks.<"$tname">.shutdown) {
      pthread_mutex_unlock(&self->tasks.<"[$task name]">.lock);
      break;
    }

    /* handle start/stop requests, wake up paused activities */
    for(id = 0; id < genom_max_activities(); id++) {
      a = &self->tasks.<"$tname">.activities.a[id].any;
      switch(a->status) {
        case ACT_VOID: case ACT_ETHER: continue;

        case ACT_INIT:
          if (a->start && self->tasks.<"$tname">.runnable) {
            /* start request, only on a tick in periodic tasks */
            a->status = ACT_RUN;
            a->state = <"${comp}">_start;
            a->start = 0;
            a->pause = 0;
          }
          if (a->stop) {
            /* stop before start: raise 'interrupted' exception, except for
             * the permanent activity for which we still go to stop */
            if (a == self->tasks.<"[$task name]">.permanent) {
              a->status = ACT_STOP;
              a->state = <"${comp}">_stop;
              a->stop = 0;
              a->pause = 0;
            } else {
              genom_interrupted_detail d;
              assert(a->interruptedby);
              strncpy(d.by, a->interruptedby, sizeof(d.by));
              d.by[sizeof(d.by)-1] = *"";
              genom_interrupted(&d, &self->tasks.<"[$task name]">.context);

              a->status = ACT_ETHER;
              a->stop = 0;
              a->exdetail = (void *)self->tasks.<"$tname">.context.raised(
                &a->state, &self->tasks.<"$tname">.context);
              sched |= GENOM_SCHED_DONE;
            }
          }
          break;

        case ACT_RUN:
          if (a->stop) {
            a->status = ACT_STOP;
            a->state = <"${comp}">_stop;
            a->stop = 0;
            a->pause = 0;
          }
          /*FALLTHROUGH*/
        case ACT_STOP:
          if (a->pause && self->tasks.<"$tname">.runnable) {
            a->pause = 0;
          }
          break;
      }
    }
    self->tasks.<"[$task name]">.runnable = 0;

    /* run activities */
    for(id = 0; id < genom_max_activities(); id++) {
      a = &self->tasks.<"$tname">.activities.a[id].any;
      if (a->status != ACT_RUN && a->status != ACT_STOP) continue;
      if (a->pause) continue;

      pthread_mutex_unlock(&self->tasks.<"[$task name]">.lock);
      e = <"genom_${comp}_[$task name]_invoke">(self, a);
      pthread_mutex_lock(&self->tasks.<"[$task name]">.lock);
      switch(e) {
        case ACT_RUN:
          if (!a->pause)
            sched |= GENOM_SCHED_CONT | GENOM_SCHED_WAKE;
          break;

        case ACT_ETHER:
          if (a->state == NULL /* exception raised */)
            a->exdetail = (void *)self->tasks.<"$tname">.context.raised(
              &a->state, &self->tasks.<"$tname">.context);
          a->status = ACT_ETHER;
          sched |= GENOM_SCHED_DONE | GENOM_SCHED_WAKE;
          break;

        case ACT_VOID: case ACT_INIT: case ACT_STOP:
          assert(!"not reached");
      }
      sched |= GENOM_SCHED_ONE;
    }

    if (sched & GENOM_SCHED_ONE)
      pthread_cond_broadcast(&self->tasks.<"[$task name]">.sync);
    if (sched & GENOM_SCHED_WAKE)
      self->tasks.<"[$task name]">.runnable = 1;
    pthread_mutex_unlock(&self->tasks.<"[$task name]">.lock);

    /* send signals */
    if (sched & GENOM_SCHED_DONE) {
      pthread_spin_lock(&self->control.rlock);
      self->control.rusage_update = tvs;
      pthread_spin_unlock(&self->control.rlock);
      genom_<"$comp">_schedule_cntrl(self);
    }
    if (sched & GENOM_SCHED_WAKE) {
<'foreach t [lsearch -inline -all -not [$component tasks] $task] {'>
      pthread_mutex_lock(&self->tasks.<"[$t name]">.lock);
      if (!self->tasks.<"[$t name]">.runnable) {
        self->tasks.<"[$t name]">.runnable = 1;
        pthread_cond_broadcast(&self->tasks.<"[$t name]">.sync);
      }
      pthread_mutex_unlock(&self->tasks.<"[$t name]">.lock);
<'}'>
    }

    /* update rusage */
<'if {![catch {$task period}]} {'>
    if (sched & GENOM_SCHED_CONT) /* at each period only */ continue;
<'}'>
    {
      double last;
      double dt;
      genom_state_rusage u = *usage;
      struct timeval rup;

<'if {[catch {$task period} period]} {'>
      struct timeval tvp = tve;
      gettimeofday(&tve, NULL);
      dt = tve.tv_sec - tvp.tv_sec + (tve.tv_usec - tvp.tv_usec)*1e-6;
<'} else {'>
      gettimeofday(&tve, NULL);
      dt = <"[$period value]">;
<'}'>

      last = tve.tv_sec - tvs.tv_sec + (tve.tv_usec - tvs.tv_usec)*1e-6;
      if (sched & GENOM_SCHED_ONE) {
        u.cycles++;

        u.timings.last = last;
        u.load.last = 100. * last / dt;

        if (u.timings.max < u.timings.last) u.timings.max = u.timings.last;
        if (u.load.max < u.load.last) u.load.max = u.load.last;
      }

      if (dt > 10.) dt = 10.;
      u.timings.avg += (last - u.timings.avg) * dt / 10.;
      u.load.avg += (100. * last - dt * u.load.avg) / 10.;

      pthread_spin_lock(&self->tasks.<"[$task name]">.rlock);
      *usage = u;
      pthread_spin_unlock(&self->tasks.<"[$task name]">.rlock);

      /* wake up control task if needed (1s timeout) */
      pthread_spin_lock(&self->control.rlock);
      rup = self->control.rusage_update;
      dt = tve.tv_sec - rup.tv_sec + (tve.tv_usec - rup.tv_usec)*1e-6;
      if (dt > 1.) {
        self->control.rusage_update = tve;
        pthread_spin_unlock(&self->control.rlock);
        genom_<"$comp">_schedule_cntrl(self);
      } else
        pthread_spin_unlock(&self->control.rlock);
    }
  } while(1);

  genom_log_info("shutting down %s exec task", "<"[$task name]">");

  if (<"$comp">_<"$tname">_mbox_size()) {
    csMboxEnd();
  }
<'if {[catch {$task period} period]} {'>
  mboxEnd(0);
<'} else {'>
  h2timerStop(timer);
  h2timerFree(timer);
<'}'>

signal_parent:
  pthread_mutex_lock(&self->tasks.<"$tname">.lock);
  self->tasks.<"[$task name]">.taskid = ERROR;
  pthread_cond_broadcast(&self->tasks.<"$tname">.sync);
  pthread_mutex_unlock(&self->tasks.<"$tname">.lock);
  return NULL;
}
